import unittest
from dbt.adapters.factory import get_adapter
from dbt.config import RuntimeConfig
from dbt.main import handle_and_check
import sys
import os
import yaml
from core.main import handle
import shutil

IS_DOCKER = os.environ.get("AM_I_IN_A_DOCKER_CONTAINER", False)


class TestArgs(object):
    def __init__(self, kwargs):
        self.which = "run"
        self.single_threaded = False
        self.profiles_dir = os.getcwd()
        self.__dict__.update(kwargs)


class DBTIntegrationTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(DBTIntegrationTest, self).__init__(*args, **kwargs)

        self.test_schema_name = "my_test_schema"
        if IS_DOCKER:
            dbt_config_dir = os.path.abspath(
                os.path.expanduser(
                    os.environ.get("DBT_CONFIG_DIR", "/home/dbt_test_user/.dbt")
                )
            )
        else:
            dbt_config_dir = os.getcwd()

        self.dbt_config_dir = dbt_config_dir
        self.dbt_profile = os.path.join(self.dbt_config_dir, "profiles.yml")

    def setUp(self):
        self.use_profile()
        self.use_default_project()
        self.set_packages()
        self.load_config()

    def tearDown(self):

        self.load_config()  # Re-set the connection
        self.run_sql('DROP SCHEMA IF EXISTS "{}" CASCADE'.format(self.test_schema_name))

        if os.path.exists("dbt_project.yml"):
            os.remove("dbt_project.yml")
        if os.path.exists("packages.yml"):
            os.remove("packages.yml")
        if os.path.exists("profiles.yml"):
            os.remove("profiles.yml")
        if os.path.exists('dbt_modules'):
            shutil.rmtree('dbt_modules')

    @property
    def project_config(self):
        return {}

    @property
    def packages_config(self):
        return None

    @property
    def test_path(self):
        path = sys.modules[self.__module__].__file__
        path = os.path.split(path)[0]
        return path

    @property
    def models_path(self, dirname="models"):
        return os.path.join(self.test_path, dirname)

    def postgres_profile(self):

        if IS_DOCKER:
            hostname = "database"
        else:
            hostname = "localhost"

        return {
            "config": {"send_anonymous_usage_stats": False},
            "test": {
                "outputs": {
                    "default2": {
                        "type": "postgres",
                        "threads": 4,
                        "host": "{}".format(hostname),
                        "port": 5432,
                        "user": "root",
                        "pass": "password",
                        "dbname": "dbt",
                        "schema": "{}".format(self.test_schema_name),
                    },
                    "noaccess": {
                        "type": "postgres",
                        "threads": 4,
                        "host": "{}".format(hostname),
                        "port": 5432,
                        "user": "noaccess",
                        "pass": "password",
                        "dbname": "dbt",
                        "schema": "{}".format(self.test_schema_name),
                    },
                },
                "target": "default2",
            },
        }

    def use_profile(self):
        if not os.path.exists(self.dbt_config_dir):
            os.makedirs(self.dbt_config_dir)

        profile_config = {}
        default_profile_config = self.postgres_profile()

        profile_config.update(default_profile_config)

        with open(self.dbt_profile, "w") as f:
            yaml.safe_dump(profile_config, f, default_flow_style=True)

        self._profile_config = profile_config

    def set_packages(self):
        if self.packages_config is not None:
            with open('packages.yml', 'w') as f:
                yaml.safe_dump(self.packages_config, f, default_flow_style=True)

    def load_config(self):
        # we've written our profile and project. Now we want to instantiate a
        # fresh adapter for the tests.
        # it's important to use a different connection handle here so
        # we don't look into an incomplete transaction
        kwargs = {"profile": None, "profiles_dir": self.dbt_config_dir, "target": None}

        config = RuntimeConfig.from_args(TestArgs(kwargs))

        adapter = get_adapter(config)
        adapter.cleanup_connections()
        self.connection = adapter.acquire_connection("__test")
        self.adapter_type = self.connection.type
        self.adapter = adapter
        self.config = config

    def run_sql(self, sql, fetch="None"):

        if sql.strip() == "":
            return

        self.load_config()
        conn = self.connection
        with conn.handle.cursor() as cursor:
            try:
                cursor.execute(sql)
                conn.handle.commit()
                if fetch == "one":
                    return cursor.fetchone()
                elif fetch == "all":
                    return cursor.fetchall()
                else:
                    return
            except BaseException as e:
                conn.handle.rollback()
                print(sql)
                print(e)
                raise e

    def use_default_project(self, overrides=None):
        # create a dbt_project.yml
        base_project_config = {
            "name": "test",
            "version": "1.0",
            "test-paths": [],
            "source-paths": [self.models_path],
            "profile": "test",
        }

        project_config = {}
        project_config.update(base_project_config)
        project_config.update(self.project_config)
        project_config.update(overrides or {})

        with open("dbt_project.yml", "w") as f:
            yaml.safe_dump(project_config, f, default_flow_style=True)

    def run_dbthelper(self, args):

        if args is None:
            args = []

        args.extend(["--profiles-dir", self.dbt_config_dir])
        results = handle(args)
        return results

    def run_dbt(self, args):
        if args is None:
            args = ["run"]

        args.extend(["--profiles-dir", self.dbt_config_dir])

        res, success = handle_and_check(args)



import networkx as nx
import dbt.loader
from dbt.config import RuntimeConfig
from dbt.logger import GLOBAL_LOGGER as logger


class ShowDependenciesTask:
    def __init__(self, args):
        self.args = args
        if self.args.command == "show_upstream":
            self.direction = "upstream"
        elif self.args.command == "show_downstream":
            self.direction = "downstream"
        else:
            raise
        self.config = RuntimeConfig.from_args(args)
        self.model_path = self.config.source_paths[0]
        self.manifest = self._get_manifest()

    def _get_manifest(self):
        manifest = dbt.loader.GraphLoader.load_all(self.config)
        return manifest

    # this traverses an arbitrary tree (parents or children) to get all ancestors or descendants
    def traverse_tree(self, node, d_tree, been_done=set()):
        tree_outputs = set(d_tree.get(node, set()))  # direct relatives
        for key in d_tree.get(node, set()):  # 2nd step relatives
            if key not in been_done:
                been_done.add(key)  # to break any circular references
                tree_outputs = tree_outputs.union(
                    self.traverse_tree(key, d_tree, been_done)
                )
        return tree_outputs

    # this reverses a parent tree to a child tree
    def get_child_dict(self, parent_dict):
        child_dict = {}
        for node in parent_dict:
            for parent in parent_dict[node]:
                if parent not in child_dict.keys():
                    child_dict[parent] = set([node])
                else:
                    child_dict[parent].add(node)
        return child_dict

    def get_node_set(self, parent_dict, focal_set):

        downstream_dict = {}  # intended to store all descendants (any generation)
        upstream_dict = {}  # intended to store all ancestry (any generation)
        node_set = set()

        # reverse parent dict to child so we can look for descendants
        child_dict = self.get_child_dict(parent_dict)  # immediate children

        # build descendant dict
        for node in child_dict:
            downstream_dict[node] = self.traverse_tree(
                node, child_dict, been_done=set()
            )

        # build ancestor dict
        for node in parent_dict:
            upstream_dict[node] = self.traverse_tree(node, parent_dict, been_done=set())

        for focal_node in focal_set:
            if self.direction == "upstream":
                node_set = upstream_dict.get(focal_node, set())
            else:
                node_set = downstream_dict.get(focal_node, set())
            node_set.add(focal_node)

        return node_set

    def dereference_model_name(self, model_name):
        for name, node in self.manifest.nodes.items():
            if node.name == model_name:
                return name

    def get_node_info(self):

        node_info_dict = {}  # intended to store direct node type
        parent_dict = {}  # intended to store parent data

        for name, node in self.manifest.nodes.items():
            d = {}
            d["name"] = name
            if node['resource_type'] == 'source':
                mat = 'source'
                schema = node['schema']
                alias = node['name']
                parents = []
            else:
                mat = node.config["materialized"]
                if len(node["fqn"]) == 3:
                    schema = node["fqn"][1]
                else:
                    schema = node["fqn"][0]

                alias = node["alias"]
                parents = node["depends_on"]["nodes"]

            d["type"] = mat
            d["alias"] = "{}.{}".format(schema, alias)

            # add the object name and type and direct parents to the dict
            node_info_dict[d["name"]] = d
            parent_dict[d["name"]] = parents

        return (parent_dict, node_info_dict)

    def build_d_graph(self, parent_dict, node_set, node_type_dict):
        G = nx.DiGraph()  # initialize directional graph object

        for node in node_set:
            G.add_node(node)
            for parent in parent_dict.get(node, set()):
                G.add_edge(parent, node)
        return G

    def display_deps(self, viz_dict):
        rev = self.direction == "downstream"
        keylist = list(viz_dict.keys())
        layers = sorted(keylist, reverse=rev)
        print("-" * 80)
        for layer in layers:
            print(" | ".join(viz_dict[layer]).center(80))
            print("-" * 80)

    def subset_dict(self, d, nodes):
        out_d = {}
        for k, v in d.items():
            if k in nodes:
                dep_set = set()
                for n in d[k]:
                    if n in nodes:
                        dep_set.add(n)
                out_d[k] = dep_set
        return out_d

    def pretty_node_name(self, name):
        return self.node_info_dict[name]["alias"]

    def run(self, args):
        parent_dict, self.node_info_dict = self.get_node_info()
        dbt_name = self.dereference_model_name(self.args.model_name)
        if not dbt_name:
            logger.info(
                dbt.ui.printer.yellow(
                    "Warning: The model argument {} does not match any models "
                    "found in this project:".format(self.args.model_name)
                )
            )
            return {}
        focal_set = set([dbt_name])

        node_set = self.get_node_set(parent_dict, focal_set)

        parent_subset_dict = self.subset_dict(parent_dict, node_set)

        G = self.build_d_graph(parent_subset_dict, node_set, self.node_info_dict)

        viz_dict = {}

        def update_viz_dict(G, current_node, level=0):
            if len(G.nodes()) == 0:
                viz_dict[0] = [self.pretty_node_name(dbt_name)]
                return
            if level in viz_dict:
                viz_dict[level].append([self.pretty_node_name(current_node)])
            else:
                viz_dict[level] = [self.pretty_node_name(current_node)]
            if self.direction == "upstream":
                if G.predecessors(current_node) == []:
                    return
                for pred in G.predecessors(current_node):
                    update_viz_dict(G, pred, level + 1)
            if self.direction == "downstream":
                if G.successors(current_node) == []:
                    return
                for pred in G.successors(current_node):
                    update_viz_dict(G, pred, level + 1)

        update_viz_dict(G, dbt_name)
        self.display_deps(viz_dict)
        return viz_dict



from dbt.config import RuntimeConfig
import dbt.adapters.factory
from dbt.node_types import NodeType

from dbt.utils import is_enabled as check_is_enabled
import dbt.loader
import dbt.ui
from dbt.logger import GLOBAL_LOGGER as logger


class CompareTask:
    def __init__(self, args):
        self.args = args
        self.config = RuntimeConfig.from_args(args)

    def _get_manifest(self):
        manifest = dbt.loader.GraphLoader.load_all(self.config)
        return manifest

    def run(self):

        # Look up all of the relations in the DB
        adapter = dbt.adapters.factory.get_adapter(self.config)
        manifest = self._get_manifest()

        schemas = set()
        model_relations = set()
        # Look up all of the relations dbt knows about
        for node in manifest.nodes.values():
            if node["resource_type"] != "source":
                schema_info = (node["database"], node["schema"])
                schemas.update([schema_info])
                node = node.to_dict()
                is_refable = (
                    node["resource_type"] in NodeType.refable()
                    or node["resource_type"] == "archive"
                )
                is_enabled = check_is_enabled(node)
                is_ephemeral = node["config"]["materialized"] == "ephemeral"
                if is_refable and is_enabled and not is_ephemeral:
                    rel = (node["schema"].lower(), node["alias"].lower())
                    model_relations.add(rel)

        db_relations = []
        for database_name, schema_name in schemas:
            db_relations.extend(adapter.list_relations(database_name, schema_name))

        database_relations = set()
        database_relations_map = dict()
        for relation in db_relations:
            relation_id = (relation.schema.lower(), relation.identifier.lower())
            database_relations_map[relation_id] = relation
            database_relations.add(relation_id)

        logger.info(
            "Comparing local models to the database catalog. " "Checking schemas:"
        )
        for database_name, schema_name in schemas:
            logger.info("- {}".format(schema_name))

        problems = database_relations - model_relations

        if len(problems) == 0:
            logger.info(
                dbt.ui.printer.green(
                    "All clear! There are no relations in the checked schemas in the database"
                    "that are not defined in dbt models."
                )
            )
        else:
            logger.info(
                dbt.ui.printer.yellow(
                    "Warning: The following relations do not match any models "
                    "found in this project:"
                )
            )

        problem_relation_list = []  # Get a list of relations to return

        for relation_id in problems:
            relation = database_relations_map[relation_id]
            problem_relation_list.append(relation)
            logger.info("{} {}".format(relation.type.upper(), relation))
            # TODO: Fix this so that it doesn't break when type is None
            # logger.info("{} {}".format(relation.type, relation))

        return problem_relation_list

    def interpret_results(self, results):
        return len(results) == 0



import os
import sys
from setuptools import find_packages, setup
from setuptools.command.install import install


package_name = "dbt-helper"
VERSION = "0.2.0"
description = """dbt-helper is a command line tool to help ease dbt development and database management"""


class VerifyVersionCommand(install):
    """
    Custom command to verify that the git tag matches our version
    https://circleci.com/blog/continuously-deploying-python-packages-to-pypi-with-circleci/
    """

    description = "verify that the git tag matches our version"

    def run(self):
        tag = os.getenv("CIRCLE_TAG")

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


setup(
    name=package_name,
    version=VERSION,
    description=description,
    author="Michael Kaminsky",
    author_email="michael@kaminsky.rocks",
    url="https://github.com/mikekaminsky/dbt-helper",
    packages=find_packages(),
    package_data={},
    test_suite="test",
    entry_points={"console_scripts": ["dbt-helper = core.main:main"]},
    scripts=[],
    install_requires=["dbt"],
    cmdclass={"verify": VerifyVersionCommand},
)



from dbt.config import RuntimeConfig

import os
import json
import subprocess


COMPILED_DIR = "compiled"
RUN_DIR = "run"
MANIFEST_FILE = "manifest.json"
DEFAULT_OPEN_COMMAND = "open"
COMPILATION_MESSAGE = "You may need to run dbt compile first."


class OpenTask:
    def __init__(self, args):
        self.args = args
        self.config = RuntimeConfig.from_args(args)
        self.model_path = self.config.source_paths[0]
        self.target_path = self.config.target_path
        self.manifest = self._get_manifest()
        self.open_command = self._get_open_command()

    def _get_manifest(self):
        """
        This subcommand uses the manifest file, whereas other subcommands import
        the manifest object from dbt (which requires the project to be parsed).
        Using the manifest file is significantly faster, so is preferred in this
        case.
        """
        manifest_path = os.path.join(self.target_path, MANIFEST_FILE)
        try:
            with open(manifest_path) as f:
                manifest = json.load(f)
            return manifest
        except IOError:
            raise Exception(
                "Could not find {} file. {}".format(MANIFEST_FILE, COMPILATION_MESSAGE)
            )

    def _get_open_command(self):
        open_command = os.environ.get(
            "DBT_HELPER_EDITOR", os.environ.get("EDITOR", DEFAULT_OPEN_COMMAND)
        )
        return open_command

    def _get_model_files(self):
        """ Return a dictionary of the form:
        {
            'source': 'models/path/to/model',
            'compiled': 'target/compiled/path/to/model',
            'run': 'target/run/path/to/model'
        }
        """

        file_names = {}
        for node in self.manifest["nodes"].values():
            if (
                node["resource_type"] == "model"
                and node["name"] == self.args.model_name
            ):
                root_path = node["root_path"]
                original_file_path = node["original_file_path"]
                package_name = node["package_name"]
                relative_path = node["path"]

                file_names["source"] = os.path.join(root_path, original_file_path)

                file_names["compiled"] = os.path.join(
                    self.target_path, COMPILED_DIR, package_name, relative_path
                )

                file_names["run"] = os.path.join(
                    self.target_path, RUN_DIR, package_name, relative_path
                )

        return file_names

    def run(self):

        model_files = self._get_model_files()

        if model_files == {}:
            raise Exception(
                "Could not find a model '{}'. {}".format(
                    self.args.model_name, COMPILATION_MESSAGE
                )
            )

        file_to_open = model_files.get(self.args.code_type)
        if not os.path.isfile(file_to_open):
            raise Exception(
                "Could not find a file '{}'. {}".format(
                    file_to_open, COMPILATION_MESSAGE
                )
            )

        result = subprocess.call(
            " ".join([self.open_command, file_to_open]), shell=True
        )
        if result == 0:
            print("Opened " + file_to_open)
        else:
            raise Exception(
                "Unsuccessfully tried to open a file using the '{}' command. "
                "You may need to update your $DBT_HELPER_EDITOR or $EDITOR "
                "environment variable.".format(self.open_command)
            )
        return result
